
#include "window.hpp"

#include <iostream>
#include <string>
#include <stdexcept>

#include <GLFW/glfw3.h>

using namespace fggl::gfx;

static void glfw_error(int code, const char* description) {
	std::cerr << "[GLFW] " << code << " " << description << std::endl;
}

static void framebuffer_resize(GLFWwindow* window, int width, int height) {
	glfwMakeContextCurrent( window );
	glViewport(0, 0, width, height);

	auto fgglWindow = reinterpret_cast<Window*>(glfwGetWindowUserPointer( window ));
	fgglWindow->framesize( width, height );
}

static void fggl_input_cursor(GLFWwindow* window, double x, double y) {
	auto fgglWindow = reinterpret_cast<Window*>(glfwGetWindowUserPointer( window ));
	Input& input = Input::instance();

	// set coords in [-1, 1]
	x = ((x / fgglWindow->width()) - 0.5f) * 2;
	y = ((y / fgglWindow->height()) - 0.5f) * 2;
	input.mousePos(x, y);
}

static void fggl_input_scroll(GLFWwindow* window, double x, double y) {
	Input& input = Input::instance();
	input.mouseScroll(x, y);
}

static void fggl_input_mouse_btn(GLFWwindow* window, int btn, int action, int mods) {
	// as we need the singleton for joysticks, might as well use it for these as well...
	Input& input = Input::instance();
	fggl::gfx::MouseButton buttonBit = (fggl::gfx::MouseButton)(1 << btn);
	input.mouseBtn( buttonBit, action == GLFW_PRESS );
}

static void fggl_activate_joystick(Input& input, int jid) {
	fggl::gfx::Joystick data;
	data.name = glfwGetJoystickName(jid);
	data.axes = glfwGetJoystickAxes(jid, &data.axisCount);
	data.buttons = glfwGetJoystickButtons(jid, &data.buttonCount);
	data.hats = glfwGetJoystickHats(jid, &data.hatCount);
	data.gamepad = glfwJoystickIsGamepad(jid);
	input.joystickConnect(jid, data);

	if ( data.gamepad ) {
		GLFWgamepadstate state;
		glfwGetGamepadState(jid, &state);

		fggl::gfx::PadState padState;
		for (int i=0; i<=GLFW_GAMEPAD_BUTTON_LAST; i++) {
			padState.buttons[i] = state.buttons[i];
		}

		for (int i=0; i<=GLFW_GAMEPAD_AXIS_LAST; i++) {
			padState.axes[i] = state.axes[i];
		}

		input.padState(jid, padState);
	}
}

static void fggl_joystick_init() {
	Input& input = Input::instance();
	for (int jid=0; jid<16; jid++) {
		if ( glfwJoystickPresent(jid) ) {
			fggl_activate_joystick(input, jid);
		}
	}
}

static void fggl_joystick_poll() {
	Input& input = Input::instance();
	for (int jid=0; jid<16; jid++) {
		if ( glfwJoystickPresent(jid) ) {
			fggl_activate_joystick(input, jid);
		}
	}
}

static void fggl_joystick(int jid, int state) {
	Input& input = Input::instance();
	if ( state == GLFW_CONNECTED ) {
		fggl_activate_joystick(input, jid);
	} else if ( state == GLFW_DISCONNECTED ) {
		input.joystickDisconnect(jid);
	}
}

Context::Context() {
	glfwInitHint(GLFW_JOYSTICK_HAT_BUTTONS, GLFW_FALSE);

	int state = glfwInit();
	if ( state == GLFW_FALSE ) {
		const char** error;
		int code = glfwGetError(error);
		throw std::runtime_error( *error );
	}

	// joysticks are global
	fggl_joystick_init();
	glfwSetJoystickCallback(fggl_joystick);
}

Context::~Context() {
	glfwTerminate();
}

void Context::pollEvents() {
	glfwPollEvents();
	fggl_joystick_poll();
}

Window::Window(Input& input) : m_window(nullptr), m_input(input) {
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, true);  

	m_window = glfwCreateWindow(1920, 1080, "main", nullptr, nullptr);
	if ( m_window == nullptr ) {
		return;
	}

	m_framesize = glm::vec2(1920, 1080);
	glfwSetWindowUserPointer(m_window, this);
	glfwSetFramebufferSizeCallback( m_window, framebuffer_resize );

	// input functions
	glfwSetScrollCallback(m_window, fggl_input_scroll);
	glfwSetCursorPosCallback(m_window, fggl_input_cursor);
	glfwSetMouseButtonCallback(m_window, fggl_input_mouse_btn);
}

Window::~Window() {
	if ( m_window != nullptr ) {
		glfwDestroyWindow(m_window);
		m_window = nullptr;
	}
}

void Window::activate() const {
	assert( m_window != nullptr );
	glfwMakeContextCurrent(m_window);
}

void Window::swap() {
	assert( m_window != nullptr );
	glfwSwapBuffers(m_window);
}
